#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <pthread.h>
#include <vector>
#include "common.h"
//#include "pthread_barrier.h"
using namespace std;

//
//  tuned constants
//
#define density 0.0005
#define mass    0.01
#define cutoff  0.01
#define min_r   (cutoff/100)
#define dt      0.0005

//
//  global variables
//
int n, n_threads;
particle_t *particles;
FILE *fsave;
pthread_barrier_t barrier;
vector<particle_t*> *bins;
int bpr, numbins;
pthread_mutex_t *binsMutex;

//
//  check that pthreads routine call was successful
//
#define P( condition ) {if( (condition) != 0 ) { printf( "\n FAILURE in %s, line %d\n", __FILE__, __LINE__ );exit( 1 );}}

// calculate particle's bin number
int binNum(particle_t &p) 
{
    return ( floor(p.x/cutoff) + bpr*floor(p.y/cutoff) );
}

//
//  This is where the action happens
//
void *thread_routine( void *pthread_id )
{
    int thread_id = *(int*)pthread_id;

    // assign particles to thread
    int particles_per_thread = (n + n_threads - 1) / n_threads;
    int first = min(  thread_id    * particles_per_thread, n );
    int last  = min( (thread_id+1) * particles_per_thread, n );

    // assign bins to thread (for clearing)
    int bins_per_thread = (numbins + n_threads - 1) / n_threads;
    int first_bin = min(  thread_id    * bins_per_thread, numbins );
    int last_bin  = min( (thread_id+1) * bins_per_thread, numbins );
    
    //
    //  simulate a number of time steps
    //
    for( int step = 0; step < NSTEPS; step++ )
    {

      // clear bins at each time step
      for (int m = first_bin; m < last_bin; m++)
      	bins[m].clear();

      pthread_barrier_wait( &barrier );
    
      // place particles in bins
      for (int i = first; i < last; i++) 
      	{
      	  int cbin = binNum(particles[i]);
	  pthread_mutex_lock( &binsMutex[cbin] );
      	  bins[cbin].push_back(particles + i );
	  pthread_mutex_unlock( &binsMutex[cbin] );
      	}

      pthread_barrier_wait( &barrier );

      //
      //  compute forces and move particles
      //
      for( int p = first; p < last; p++ )
	  {
	    particles[p].ax = particles[p].ay = 0;
          
	    // find current particle's bin, handle boundaries
	    int cbin = binNum( particles[p] );
	    int lowi = -1, highi = 1, lowj = -1, highj = 1;
	    if (cbin < bpr)
	      lowj = 0;
	    if (cbin % bpr == 0)
	      lowi = 0;
	    if (cbin % bpr == (bpr-1))
	      highi = 0;
	    if (cbin >= bpr*(bpr-1))
	      highj = 0;

	    // apply nearby forces
	    for (int i = lowi; i <= highi; i++)
	      for (int j = lowj; j <= highj; j++)   
		{
		  int nbin = cbin + i + bpr*j;
		  for (int k = 0; k < bins[nbin].size(); k++)
		      apply_force( particles[p], *bins[nbin][k] );
		}
	  }

      pthread_barrier_wait( &barrier );

      // move particle
      for( int p = first; p < last; p++ )
	move( particles[p] );
      
      //
      //  save if necessary
      //
      if( thread_id == 0 && fsave && (step%SAVEFREQ) == 0 )
	save( fsave, n, particles );
    }
    
    return NULL;
}

//
//  benchmarking program
//
int main( int argc, char **argv )
{    
    //
    //  process command line
    //
    if( find_option( argc, argv, "-h" ) >= 0 )
    {
        printf( "Options:\n" );
        printf( "-h to see this help\n" );
        printf( "-n <int> to set the number of particles\n" );
        printf( "-p <int> to set the number of threads\n" );
        printf( "-o <filename> to specify the output file name\n" );
        return 0;
    }
    
    n = read_int( argc, argv, "-n", 1000 );
    n_threads = read_int( argc, argv, "-p", 2 );
    char *savename = read_string( argc, argv, "-o", NULL );
    
    //
    //  allocate resources
    //
    fsave = savename ? fopen( savename, "w" ) : NULL;

    particles = (particle_t*) malloc( n * sizeof(particle_t) );
    set_size( n );
    init_particles( n, particles );

    pthread_attr_t attr;
    P( pthread_attr_init( &attr ) );
    P( pthread_barrier_init( &barrier, NULL, n_threads ) );

    int *thread_ids = (int *) malloc( n_threads * sizeof( int ) );
    for( int i = 0; i < n_threads; i++ ) 
        thread_ids[i] = i;

    pthread_t *threads = (pthread_t *) malloc( n_threads * sizeof( pthread_t ) );

    // create spatial bins (of size cutoff by cutoff)
    double size = sqrt( density*n );
    bpr = ceil(size/cutoff);
    numbins = bpr*bpr;
    bins = new vector<particle_t*>[numbins];

    // initialize bin mutexes
    binsMutex = new pthread_mutex_t[numbins];
    for (int m = 0; m < numbins; m++)
      pthread_mutex_init(&binsMutex[m],NULL);
    
    //
    //  do the parallel work
    //
    double simulation_time = read_timer( );
    for( int i = 1; i < n_threads; i++ ) 
        P( pthread_create( &threads[i], &attr, thread_routine, &thread_ids[i] ) );
    
    thread_routine( &thread_ids[0] );
    
    for( int i = 1; i < n_threads; i++ ) 
        P( pthread_join( threads[i], NULL ) );
    simulation_time = read_timer( ) - simulation_time;
    
    printf( "n = %d, n_threads = %d, simulation time = %g seconds\n", n, n_threads, simulation_time );
    
    //
    //  release resources
    //
    P( pthread_barrier_destroy( &barrier ) );
    P( pthread_attr_destroy( &attr ) );
    free( thread_ids );
    free( threads );
    free( particles );
    if( fsave )
        fclose( fsave );
    
    return 0;
}
