#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "common.h"

#include <vector>

using namespace std;

// the offset indicates...
// 012
// 345
// 678
int offset_x[] = {-1, -1, -1,  0,  0,  0, +1, +1, +1};
int offset_y[] = {-1,  0, +1, -1,  0, +1, -1,  0, +1};

// stuff
int step;
int n_proc, rank;
int n;
int bs;  // bin size
int lbs; // partitioned bin size
double cutoff;

// arrays
#define b(x, y, n) (bins[(x)+(y)*(n)])
#define mq(m, t, n) (message_queue[(m)+(t)*(n)])
#define mc(t) (message_count[(t)])

#define particles_prev (*particles_p)
#define particles_next (*particles_n)

bool flag = true;
vector<particle_t> particles1;
vector<particle_t> particles2;
vector<particle_t>* particles_p;
vector<particle_t>* particles_n;
vector<particle_t*>* bins;

typedef struct
{
  double x;
  double y;
  double vx;
  double vy;
} message_particle_t;

int* message_count;
message_particle_t* message_queue;
message_particle_t* message_recv;

MPI_Datatype PARTICLE;

int* x_start;
int* x_end;

void get_boundaries(int proc, int* x_start, int* x_end)
{
  *x_start = max(proc*lbs-1, 0);
  *x_end = min((proc+1)*lbs+1, bs);
}

void assign_bins()
{
  for (int p=0; p<particles_prev.size(); p++)
  {
    int x = (int)(particles_prev[p].x/cutoff);
    int y = (int)(particles_prev[p].y/cutoff);

    particles_prev[p].ax = particles_prev[p].ay = 0;

    b(x, y, bs).push_back(&particles_prev[p]);
  }
}

void clean_bins()
{
  for (int x=0; x<bs; x++)
  {
    for (int y=0; y<bs; y++)
    {
      b(x, y, bs).clear();
    }
  }
}

void clear_mqs()
{
  memset(message_count, 0, sizeof(int)*n_proc);
}

void bundle_message(particle_t& particle)
{
  int x = (int)(particle.x/cutoff);

  for (int t=0; t<n_proc; t++)
  {
    if (x_start[t] <= x && x < x_end[t])
    {
      //printf("rank=%d step=%d x=%d t=%d\n", rank, step, x, t);

      if (t == rank)
      {
        particles_next.push_back(particle);
      }
      else
      {
        memcpy(&mq(mc(t)++, t, n), &particle, sizeof(message_particle_t));
      }
    }
  }
}

void bundle_all(int n, particle_t* particles)
{
  for (int p=0; p<n; p++)
  {
    bundle_message(particles[p]);
  }
}

void send_all_messages()
{
  for (int t=0; t<n_proc; t++)
  {
    if (t == rank)
    {
      continue;
    }

    //printf("rank=%d step=%d send_messages to rank %d count=%d\n", rank, step, t, mc(t));
    MPI_Send(&mq(0, t, n), mc(t), PARTICLE, t, 0, MPI_COMM_WORLD);
  }
}

void receive_rank0_messages()
{
  MPI_Status status;
  int count;

  MPI_Recv(message_recv, n, PARTICLE, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Get_count(&status, PARTICLE, &count);

  //printf("rank=%d step=%d receive_messages from rank %d count=%d\n", rank, step, 0, count);

  particle_t p;

  for (int i=0; i<count; i++)
  {
    memcpy(&p, &message_recv[i], sizeof(message_particle_t));
    p.ax = p.ay = 0;
    particles_next.push_back(p);
    //printf("rank=%d step=%d receive_messages i=%d p.x=%6.3f p.y=%6.3f\n", rank, step, i, p.x, p.y);
  }
}

void send_messages()
{
  if (n_proc == 1) return;

  if (rank == 0)
  {
    MPI_Send(&mq(0, 1, n), mc(1), PARTICLE, 1, 0, MPI_COMM_WORLD);
  }
  else if (rank == n_proc-1)
  {
    MPI_Send(&mq(0, n_proc-2, n), mc(n_proc-2), PARTICLE, n_proc-2, 0, MPI_COMM_WORLD);
  }
  else
  {
    MPI_Send(&mq(0, rank-1, n), mc(rank-1), PARTICLE, rank-1, 0, MPI_COMM_WORLD);
    MPI_Send(&mq(0, rank+1, n), mc(rank+1), PARTICLE, rank+1, 0, MPI_COMM_WORLD);
  }
}

void receive_message_rank(int src)
{
  MPI_Status status;
  int count;

  MPI_Recv(message_recv, n, PARTICLE, src, 0, MPI_COMM_WORLD, &status);
  MPI_Get_count(&status, PARTICLE, &count);

  //printf("rank=%d step=%d receive_messages from rank %d count=%d\n", rank, step, src, count);

  particle_t p;

  for (int i=0; i<count; i++)
  {
    memcpy(&p, &message_recv[i], sizeof(message_particle_t));
    p.ax = p.ay = 0;
    particles_next.push_back(p);
    //printf("rank=%d step=%d receive_messages i=%d p.x=%6.3f p.y=%6.3f\n", rank, step, i, p.x, p.y);
  }
}

void receive_messages()
{
  if (n_proc == 1) return;

  if (rank == 0)
  {
    receive_message_rank(1);
  }
  else if (rank == n_proc-1)
  {
    receive_message_rank(n_proc-2);
  }
  else
  {
    receive_message_rank(rank-1);
    receive_message_rank(rank+1);
  }
}

//
//  benchmarking program
//
int main(int argc, char **argv)
{  
  //
  //  process command line parameters
  //
  if (find_option(argc, argv, "-h") >= 0)
  {
    printf("Options:\n");
    printf("-h to see this help\n");
    printf("-n <int> to set the number of particles\n");
    printf("-o <filename> to specify the output file name\n");
    return 0;
  }
  
  n = read_int(argc, argv, "-n", 1000);
  char *savename = read_string(argc, argv, "-o", NULL);
  
  //
  //  set up MPI
  //
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &n_proc);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  
  //
  //  allocate generic resources
  //
  FILE *fsave = savename && rank == 0 ? fopen(savename, "w") : NULL;
  
  MPI_Type_contiguous(4, MPI_DOUBLE, &PARTICLE);
  MPI_Type_commit(&PARTICLE);
  
  //
  //  initialize and distribute the particles (that's fine to leave it unoptimized)
  //
  set_size(n);
  bs = get_binsize();
  lbs = bs / n_proc + 1;
  cutoff = get_cutoff();

  bins = new vector<particle_t*> [bs*bs];
  x_start = new int [n_proc];
  x_end = new int [n_proc];
  message_count = new int [n_proc];
  message_queue = new message_particle_t [n_proc*n];
  message_recv = new message_particle_t [n];

  for (int t=0; t<n_proc; t++)
  {
    get_boundaries(t, &x_start[t], &x_end[t]);
  }

  particles_p = &particles1;
  particles_n = &particles2;

  if (rank == 0)
  {
    particle_t *iparticles = new particle_t [n];
    init_particles(n, iparticles);
    clear_mqs();
    bundle_all(n, iparticles);
    send_all_messages();
    delete [] iparticles;
  }
  else
  {
    receive_rank0_messages();
  }

  int g_start = min(rank*lbs, bs);
  int g_end = min((rank+1)*lbs, bs);

  //
  //  simulate a number of time steps
  //
  double simulation_time = read_timer();
  double message_time = 0;
  for (step = 0; step < NSTEPS; step++)
  {
    if (flag)
    {
      particles_p = &particles2;
      particles_n = &particles1;
    }
    else
    {
      particles_p = &particles1;
      particles_n = &particles2;
    }

    flag = !flag;
    particles_next.clear();
    clean_bins();
    assign_bins();

    for (int bx=g_start; bx<g_end; bx++)
    {
      for (int by=0; by<bs; by++)
      {
        if (b(bx, by, bs).size() != 0)
        {
          for (int o=0; o<9; o++)
          {
            int nx = bx+offset_x[o];
            int ny = by+offset_y[o];

            if (nx < 0 || nx >= bs) continue;
            if (ny < 0 || ny >= bs) continue;

            if (b(nx, ny, bs).size() != 0)
            {
              for (int i=0; i<b(bx, by, bs).size(); i++)
              {
                for (int j=0; j<b(nx, ny, bs).size(); j++)
                {
                  apply_force(*(b(bx, by, bs)[i]), *(b(nx, ny, bs)[j]));
                }
              }
            }
          }
        }
      }
    }

    //
    //  move particles
    //

    clear_mqs();

    for (int bx=g_start; bx<g_end; bx++)
    {
      for (int by=0; by<bs; by++)
      {
        for (int i=0; i<b(bx, by, bs).size(); i++)
        {
          move(*(b(bx, by, bs)[i]));
          bundle_message(*(b(bx, by, bs)[i]));
        }
      }
    }

    //
    //  save current step if necessary (slightly different semantics than in other codes)
    //
    if (fsave && (step%SAVEFREQ) == 0)
    {
      savevector(fsave, n, particles_prev);
    }

    double start_time = read_timer();
    send_messages();
    receive_messages();
    message_time += (read_timer() - start_time);
  }
  simulation_time = read_timer() - simulation_time;
  
  {
    if (flag)
    {
      particles_p = &particles2;
      particles_n = &particles1;
    }
    else
    {
      particles_p = &particles1;
      particles_n = &particles2;
    }

    flag = !flag;
    particles_next.clear();
    clean_bins();
    assign_bins();

    int count = 0;

    for (int bx=g_start; bx<g_end; bx++)
    {
      for (int by=0; by<bs; by++)
      {
        for (int i=0; i<b(bx, by, bs).size(); i++)
        {
          printf("final rank=%d count=%d %f %f\n", rank, count++, b(bx, by, bs)[i]->x, b(bx, by, bs)[i]->y);
        }
      }
    }
  }
  if (rank == 0)
  {
    printf("n = %d, n_procs = %d, simulation time = %g s\n", n, n_proc, simulation_time);
  }

  printf("rank=%d message_time = %g s\n", rank, message_time);
  
  //
  //  release resources
  //

  if (fsave)
    fclose(fsave);
  
  MPI_Finalize();
  
  return 0;
}
