#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <omp.h>
#include "common.h"

// the offset indicates...
// 012
// 345
// 678
int offset_x[] = {-1, -1, -1,  0,  0,  0, +1, +1, +1};
int offset_y[] = {-1,  0, +1, -1,  0, +1, -1,  0, +1};

// stuff
int bs; // bin size
double cutoff;

class grid
{
public:
  grid(int _bs)
  {
    abs = _bs;
    bs = _bs;
    np = sqrt(bs);

    bins = new int [bs*bs*np];
    counts = new int [bs*bs];
    locks = new omp_lock_t [bs*bs];

    posix_memalign((void**)&bins, 4096, bs*bs*np*sizeof(int));
    posix_memalign((void**)&counts, 4096, bs*bs*sizeof(int));
    posix_memalign((void**)&locks, 4096, bs*bs*sizeof(omp_lock_t));
  }

  ~grid()
  {
    delete [] bins;
    delete [] counts;
    delete [] locks;
  }

  inline void init()
  {
    for (int x=0; x<abs; x++)
    {
      for (int y=0; y<abs; y++)
      {
        omp_init_lock(&locks[x*bs+y]);
      }
    }
  }

  inline void clear(int x, int y)
  {
    counts[x*bs+y] = 0;
  }

  inline int count(int x, int y)
  {
    return counts[x*bs+y];
  }

  inline int* list(int x, int y)
  {
    return &bins[np*(x*bs+y)];
  }

  inline void add(int x, int y, int p)
  {
    int idx = x*bs+y;

    omp_set_lock(&locks[idx]);
    int count = counts[idx]++;
    omp_unset_lock(&locks[idx]);

    bins[np*idx+count] = p;
  }

  int abs; // actual bin size
  int bs; // bin size
  int np; // numner of particles per bin
  int* restrict bins;
  int* restrict counts;
  omp_lock_t* restrict locks;
};


//
//  benchmarking program
//
int main(int argc, char **argv)
{    
  if (find_option(argc, argv, "-h") >= 0)
  {
    printf("Options:\n");
    printf("-h to see this help\n");
    printf("-n <int> to set the number of particles\n");
    printf("-p <int> to set the number of threads\n");
    printf("-o <filename> to specify the output file name\n");
    return 0;
  }

  int n = read_int(argc, argv, "-n", 1000);
  int nthreads = read_int(argc, argv, "-p", 2);
  char *savename = read_string(argc, argv, "-o", NULL);

  FILE *fsave = savename ? fopen(savename, "w") : NULL;
  particle_t* restrict particles = (particle_t*) malloc(n * sizeof(particle_t));
  posix_memalign((void**)&particles, 4096, n*sizeof(particle_t));

  set_size(n);
  init_particles(n, particles);

  bs = get_binsize();
  cutoff = get_cutoff();

  //
  //  simulate a number of time steps
  //
  double simulation_time = read_timer();

  grid g1(bs);
  grid g2(bs);

  #pragma omp parallel num_threads(nthreads)
  {
    bool flag = true;

    grid* gp = &g1;
    grid* gn = &g2;

    int nt = omp_get_num_threads();
    int thread_id = omp_get_thread_num();
    int lbs = bs / nt + 1;

    int particles_per_thread = (n + nthreads - 1) / nthreads;
    int first = min( thread_id    * particles_per_thread, n);
    int last  = min((thread_id+1) * particles_per_thread, n);

    int x_start = min(thread_id*lbs, bs);
    int x_end = min((thread_id+1)*lbs, bs);

    for (int bx=x_start; bx<x_end; bx++)
    {
      for (int by=0; by<bs; by++)
      {
        gp->clear(bx, by);
        gn->clear(bx, by);
      }
    }

    #pragma omp master
    {
      gp->init();
      gn->init();
      printf("nthreads = %d\n", omp_get_num_threads());

      for (int i=0; i<n; i++)
      {
        int x = (int)(particles[i].x/cutoff);
        int y = (int)(particles[i].y/cutoff);

        gp->add(x, y, i);
      }
    }

    for (int step = 0; step < NSTEPS; step++)
    {
      #pragma omp barrier

      for (int bx=x_start; bx<x_end; bx++)
      {
        for (int by=0; by<bs; by++)
        {
          int count1 = gp->count(bx, by);

          if (count1)
          {
            int* restrict p1 = gp->list(bx, by);

            for (int o=0; o<9; o++)
            {
              int nx = bx+offset_x[o];
              int ny = by+offset_y[o];

              if (nx < 0 || nx >= bs) continue;
              if (ny < 0 || ny >= bs) continue;

              int count2 = gp->count(nx, ny);

              if (count2)
              {
                int* restrict p2 = gp->list(nx, ny);

                for (int i=0; i<count1; i++)
                {
                  for (int j=0; j<count2; j++)
                  {
                    apply_force(particles[p1[i]], particles[p2[j]]);
                  }
                }
              }
            }
          }
        }
      }

      #pragma omp barrier

      for (int bx=x_start; bx<x_end; bx++)
      {
        for (int by=0; by<bs; by++)
        {
          int count = gp->count(bx, by);

          if (count)
          {
            int* restrict p = gp->list(bx, by);

            for (int i=0; i<count; i++)
            {
              int idx = p[i];

              move(particles[idx]);

              particles[idx].ax = 0;
              particles[idx].ay = 0;

              int x = (int)(particles[idx].x/cutoff);
              int y = (int)(particles[idx].y/cutoff);

              gn->add(x, y, idx);
            }

            gp->clear(bx, by);
          }
        }
      }

      if (flag)
      {
        flag = false;
        gp = &g2;
        gn = &g1;
      }
      else
      {
        flag = true;
        gp = &g1;
        gn = &g2;
      }

      //
      //  save if necessary
      //
      if (fsave)
      {
        #pragma omp barrier

        if ((step%SAVEFREQ) == 0)
        {
          #pragma omp master
          {
            save(fsave, n, particles);
          }
        }
      }
    }
  }
  simulation_time = read_timer() - simulation_time;

  printf("n = %d, simulation time = %g seconds\n", n, simulation_time);

  free(particles);
  if (fsave)
    fclose(fsave);

  return 0;
}
